@@ -24,20 +24,19 @@
 //! ```rust
 //! use guff::{GaloisField, F4, new_gf4};
 //! 
-//! fn main() {
-//!     // Create a GF(2<sup>4</sup>) field from a struct
-//!     // * `19` is the field's irreducible polynomial
-//!     // * `3` is the same value with bit 0x10 removed
-//!     let f = guff::F4 { full : 19, compact : 3 };
-//!     
-//!     assert_eq!(f.pow(5,3), f.mul(5,f.mul(5,5)) );
+//! // Create a GF(2<sup>4</sup>) field from a struct
+//! // * `19` is the field's irreducible polynomial
+//! // * `3` is the same value with bit 0x10 removed
+//! let f = guff::F4 { full : 19, compact : 3 };
 //! 
-//!     // The same, but using a constructor as syntactic sugar
-//!     let f2 = guff::new_gf4(19, 3);
+//! assert_eq!(f.pow(5,3), f.mul(5,f.mul(5,5)) );
+//! 
+//! // The same, but using a constructor as syntactic sugar
+//! let f2 = guff::new_gf4(19, 3);
 //!
-//!     assert_eq!(f2.pow(5,3), f2.mul(5,f2.mul(5,5)) );
-//!     assert_eq!(f.pow(5,3), f2.pow(5,3));
-//! }
+//! assert_eq!(f2.pow(5,3), f2.mul(5,f2.mul(5,5)) );
+//! assert_eq!(f.pow(5,3), f2.pow(5,3));
+//! 
 //! 
 //! ```
 //!
@@ -74,9 +73,11 @@
 //! > Noun *guff* - unacceptable behavior (especially ludicrously
 //!   false statements)
 
-use num_traits;
+//use num_traits;
 use num::{PrimInt,One,Zero};
 
+pub mod good;
+
 /// A typing trait meant to map to a primitive unsigned integer type
 /// such as u8, u16 or u32.
 pub trait ElementStore : 'static + Copy
@@ -112,6 +113,8 @@ pub trait GaloisField {
     /// * storing the result of a non-modular (overflowing) multiply
     /// of two field elements
     type EE : ElementStore; // where Self::EE : From<Self::E>;
+    /// As EE, but signed
+    type SEE : ElementStore; // where Self::EE : From<Self::E>;
 
     /// Size of the field in bits, eg GF(2<sup>8</sup>) &rarr; 8
     const ORDER      : u16;
@@ -198,14 +201,16 @@ pub trait GaloisField {
 	let mut z : Self::E = zero;
 	let mut g : Self::E = one;
 	let mut t : Self::E;	// always set before being used
+	let mask = self.field_mask();
 	// special cases of 1/1 and 1/0
 	if a == zero || a == one { return a }
 	// unroll first loop iteration (knowing initial i >= 0)
 	// rustc can't determine that i is always positive here,
 	// though, so we have to try_into()
-	let mut i : u32 = 1 + v.leading_zeros();
-	u = u ^ v << i as usize;
-	z = z ^ g << i as usize;
+	let fixup = if self.order() == 4 { 4 } else { 0 };
+	let mut i : u32 = 1 + v.leading_zeros() - fixup;
+	u = (u ^ v << i as usize) & mask;
+	z = (z ^ g << i as usize) & mask;
 	while u != one {
 	    // in C: i=size_in_bits(u) - size_in_bits(v)
 	    // size_in_bits is order - leading zeroes
@@ -222,8 +227,8 @@ pub trait GaloisField {
 		//		    i = v.leading_zeros() - u.leading_zeros();
 	    }
 	    i = v.leading_zeros() - u.leading_zeros();
-	    u = u ^ v << i as usize;
-	    z = z ^ g << i as usize;
+	    u = (u ^ v << i as usize) & mask;
+	    z = (z ^ g << i as usize) & mask;
 	}
 	z
     }
@@ -419,6 +424,7 @@ pub struct F32 { pub full : u64, pub compact : u32 }
 impl GaloisField for F4 {
     type E = u8;
     type EE = u8;
+    type SEE = i8;
 
     // we have to redeclare types for constants
     const ORDER      : u16 = 4;
@@ -436,12 +442,13 @@ impl GaloisField for F4 {
 /// Create a new GF(2<sup>4</sup>) field with a supplied field
 /// polynomial (using the default implementation)
 pub fn new_gf4(full : u8, compact : u8) -> F4  {
-    F4 {full : full, compact : compact}
+    F4 {full, compact}
 }
 
 impl GaloisField for F8 {
     type E = u8;
     type EE = u16;
+    type SEE = i16;
 
     // we have to redeclare types for constants
     const ORDER      : u16 = 8;
@@ -459,12 +466,13 @@ impl GaloisField for F8 {
 /// Create a new GF(2<sup>8</sup>) field with a supplied field
 /// polynomial (using the default implementation)
 pub fn new_gf8(full : u16, compact : u8) -> F8  {
-    F8 {full : full, compact : compact}
+    F8 {full, compact}
 }
 
 impl GaloisField for F16 {
     type E = u16;
     type EE = u32;
+    type SEE = i32;
 
     // we have to redeclare types for constants
     const ORDER      : u16 = 8;
@@ -482,12 +490,13 @@ impl GaloisField for F16 {
 /// Create a new GF(2<sup>16</sup>) field with a supplied field
 /// polynomial (using the default implementation)
 pub fn new_gf16(full : u32, compact : u16) -> F16  {
-    F16 {full : full, compact : compact}
+    F16 { full, compact }
 }
 
 impl GaloisField for F32 {
     type E = u32;
     type EE = u64;
+    type SEE = i64;
 
     // we have to redeclare types for constants
     const ORDER      : u16 = 32;
@@ -505,7 +514,7 @@ impl GaloisField for F32 {
 /// Create a new GF(2<sup>32</sup>) field with a supplied field
 /// polynomial (using the default implementation)
 pub fn new_gf32(full : u64, compact : u32) -> F32  {
-    F32 {full : full, compact : compact}
+    F32 { full, compact }
 }
 
 
@@ -941,4 +950,13 @@ mod tests {
 	}
     }
 
+    #[test]
+    fn test_gf8_inverses() {
+	let f = new_gf8(0x11b, 0x1b);
+	for a in 0..=255 {
+	    let inv = f.inv(a);
+	    let invinv = f.inv(inv);
+	    assert_eq!(a, invinv);
+	}
+    }
 }
